from __future__ import annotations

import re
from typing import TYPE_CHECKING, Any

import numpy as np
import pytest
from pydantic import ValidationError
from pydantic_zarr.v2 import AnyArraySpec, AnyGroupSpec, ArraySpec, GroupSpec

from ome_zarr_models.common.coordinate_transformations import (
    _build_transforms,
)
from ome_zarr_models.v04.axes import Axis
from ome_zarr_models.v04.coordinate_transformations import (
    VectorScale,
    VectorTranslation,
)
from ome_zarr_models.v04.image import Image, ImageAttrs
from ome_zarr_models.v04.multiscales import (
    Dataset,
    Multiscale,
)

if TYPE_CHECKING:
    from zarr.abc.store import Store

DEFAULT_UNITS_MAP = {"space": "meter", "time": "second"}


@pytest.fixture
def default_multiscale() -> Multiscale:
    """
    Return a valid Multiscale object.
    """
    axes = (
        Axis(name="c", type="channel", unit=None),
        Axis(name="z", type="space", unit="meter"),
        Axis(name="x", type="space", unit="meter"),
        Axis(name="y", type="space", unit="meter"),
    )
    rank = len(axes)
    transforms_top = _build_transforms(scale=(1,) * rank, translation=None)
    transforms_dset = _build_transforms(scale=(1,) * rank, translation=(0,) * rank)
    num_datasets = 3
    datasets = tuple(
        Dataset(path=f"path{idx}", coordinateTransformations=transforms_dset)
        for idx in range(num_datasets)
    )

    multi = Multiscale(
        axes=axes,
        datasets=datasets,
        coordinateTransformations=transforms_top,
    )
    return multi


def test_immutable(default_multiscale: Multiscale) -> None:
    """
    Check that models are immutable.
    """
    with pytest.raises(ValidationError, match="Instance is frozen"):
        default_multiscale.axes[0].name = "new_name"  # type: ignore[misc]


def test_multiscale_unique_axis_names() -> None:
    # TODO: is unique names actually part of the spec???
    axes = (
        Axis(name="x", type="space", unit="meter"),
        Axis(name="x", type="space", unit="meter"),
    )
    rank = len(axes)
    datasets = (Dataset.build(path="path", scale=(1,) * rank, translation=(0,) * rank),)

    with pytest.raises(ValidationError, match="Axis names must be unique."):
        Multiscale(
            axes=axes,
            datasets=datasets,
            coordinateTransformations=_build_transforms(scale=(1, 1), translation=None),
        )


@pytest.mark.parametrize(
    "axis_types",
    [
        ("space", "space", "channel"),
    ],
)
def test_multiscale_space_axes_last(axis_types: list[str]) -> None:
    """
    Error if the last axes isn't 'space'.

    > ... the entries MUST be ordered by "type" where the
    > "time" axis must come first (if present), followed by the "channel" or
    > custom axis (if present) and the axes of type "space".
    """
    units_map = {"space": "meter", "time": "second"}
    axes = tuple(
        Axis(name=str(idx), type=t, unit=units_map.get(t))
        for idx, t in enumerate(axis_types)
    )
    rank = len(axes)
    datasets = (Dataset.build(path="path", scale=(1,) * rank, translation=(0,) * rank),)
    # TODO: make some axis-specific exceptions
    with pytest.raises(
        ValidationError, match="All space axes must be at the end of the axes list."
    ):
        Multiscale(
            axes=axes,
            datasets=datasets,
            coordinateTransformations=_build_transforms(
                scale=(1,) * rank, translation=None
            ),
        )


@pytest.mark.parametrize(
    "axis_types",
    [
        ("channel", "time", "space", "space"),
    ],
)
def test_axes_order(axis_types: list[str]) -> None:
    """
    If 'time' is present, it must be first

    > ... the entries MUST be ordered by "type" where the
    > "time" axis must come first (if present), followed by the "channel" or
    > custom axis (if present) and the axes of type "space".
    """
    axes = tuple(
        Axis(name=str(idx), type=t, unit=DEFAULT_UNITS_MAP.get(t))
        for idx, t in enumerate(axis_types)
    )
    rank = len(axes)
    datasets = (Dataset.build(path="path", scale=(1,) * rank, translation=(0,) * rank),)
    with pytest.raises(
        ValidationError, match="Time axis must be at the beginning of axis list"
    ):
        Multiscale(
            axes=axes,
            datasets=datasets,
            coordinateTransformations=_build_transforms(
                scale=(1,) * rank, translation=None
            ),
        )


@pytest.mark.parametrize("num_axes", [0, 1, 6, 7])
def test_multiscale_axis_length(num_axes: int) -> None:
    """
    > The length of "axes" must be between 2 and 5...
    """
    rank = num_axes
    axes = tuple(
        Axis(name=str(idx), type="space", unit="meter") for idx in range(num_axes)
    )
    datasets = (Dataset.build(path="path", scale=(1,) * rank, translation=(0,) * rank),)
    with pytest.raises(ValidationError, match=r"Length of axes \([0-9]+\) not valid"):
        Multiscale(
            axes=axes,
            datasets=datasets,
            coordinateTransformations=_build_transforms(
                scale=(1,) * rank, translation=None
            ),
        )


def test_invalid_dataset_dimensions() -> None:
    """
    > Each "datasets" dictionary MUST have the same number of dimensions...
    """
    datasets = [
        Dataset.build(path="path", scale=(1,) * rank, translation=(0,) * rank)
        for rank in [2, 3]
    ]
    axes = tuple(Axis(name=str(idx), type="space", unit="meter") for idx in range(3))
    with pytest.raises(
        ValidationError,
        match=(
            "The length of axes does not match the dimensionality "
            "of the scale transform"
        ),
    ):
        Multiscale(
            axes=axes,
            datasets=datasets,
        )


@pytest.mark.parametrize(
    "scale, translation", [((1, 1), (1, 1, 1)), ((1, 1, 1), (1, 1))]
)
def test_transform_invalid_ndims(
    scale: tuple[int, ...], translation: tuple[int, ...]
) -> None:
    """
    Make sure dimensions of scale/translation transforms match.
    """
    with pytest.raises(
        ValidationError,
        match="The transforms have inconsistent dimensionality.",
    ):
        Dataset.build(path="foo", scale=scale, translation=translation)


@pytest.mark.parametrize(
    "transforms",
    [
        (
            VectorScale.build((1, 1, 1)),
            VectorTranslation.build((1, 1, 1)),
            VectorTranslation.build((1, 1, 1)),
        ),
        (VectorScale.build((1, 1, 1)),) * 5,
    ],
)
def test_transform_invalid_length(
    transforms: tuple[Any, ...],
) -> None:
    """
    Error if there's the wrong number of transforms.
    """
    with pytest.raises(
        ValidationError,
        match=re.escape(f"Length of transforms ({len(transforms)}) not valid"),
    ):
        Dataset(path="foo", coordinateTransformations=transforms)


@pytest.mark.parametrize(
    "transforms",
    [
        (VectorTranslation.build((1, 1, 1)),) * 2,
        (
            VectorTranslation.build((1, 1, 1)),
            VectorScale.build((1, 1, 1)),
        ),
    ],
)
def test_transform_invalid_first_element(
    transforms: tuple[Any, Any],
) -> None:
    """
    Make sure first transform element is a scale.
    """
    with pytest.raises(
        ValidationError,
        match="The first element of `coordinateTransformations` "
        "must be a scale transform",
    ):
        Dataset(path="foo", coordinateTransformations=transforms)


@pytest.mark.parametrize(
    "transforms",
    (
        (
            VectorScale.build((1, 1, 1)),
            VectorScale.build((1, 1, 1)),
        ),
    ),
)
def test_transform_invalid_second_element(
    transforms: tuple[VectorScale, VectorScale],
) -> None:
    """
    Make sure second transform is a translation.
    """
    with pytest.raises(
        ValidationError,
        match="The second element of `coordinateTransformations` "
        "must be a translation transform",
    ):
        Dataset(path="foo", coordinateTransformations=transforms)


def test_validate_axes_top_transforms() -> None:
    """
    Test that the number of axes must match the dimensionality of the
    top-level coordinateTransformations.
    """
    axes_rank = 3
    tforms_rank = 2
    msg_expect = (
        f"The length of axes does not match the dimensionality of "
        f"the scale transform in coordinateTransformations. "
        f"Got {axes_rank} axes, but the scale transform has "
        f"dimensionality {tforms_rank}"
    )
    with pytest.raises(
        ValidationError,
        match=msg_expect,
    ):
        Multiscale(
            name="foo",
            axes=[Axis(name=str(idx), type="space") for idx in range(axes_rank)],
            datasets=(
                Dataset.build(
                    path="foo", scale=(1,) * axes_rank, translation=(0,) * axes_rank
                ),
            ),
            coordinateTransformations=_build_transforms(
                scale=(1,) * tforms_rank, translation=None
            ),
        )


def test_validate_axes_dset_transforms() -> None:
    """
    Test that the number of axes must match the dimensionality of the
    per-dataset coordinateTransformations
    """
    axes_rank = 3
    tforms_rank = 2
    axes = [Axis(name=str(idx), type="space") for idx in range(axes_rank)]

    msg_expect = (
        f"The length of axes does not match the dimensionality of "
        f"the scale transform in datasets[0].coordinateTransformations. "
        f"Got {axes_rank} axes, but the scale transform has "
        f"dimensionality {tforms_rank}"
    )

    with pytest.raises(
        ValidationError,
        match=re.escape(msg_expect),
    ):
        Multiscale(
            name="foo",
            axes=axes,
            datasets=[
                Dataset.build(
                    path="foo", scale=(1,) * tforms_rank, translation=(0,) * tforms_rank
                )
            ],
            coordinateTransformations=_build_transforms(
                scale=(1,) * axes_rank, translation=None
            ),
        )


def test_ordered_multiscales() -> None:
    """
    > The "path"s MUST be ordered from largest (i.e. highest resolution) to smallest.
    """
    axes = (
        Axis(name="c", type="channel", unit=None),
        Axis(name="z", type="space", unit="meter"),
        Axis(name="x", type="space", unit="meter"),
        Axis(name="y", type="space", unit="meter"),
    )
    datasets = (
        Dataset(
            path="0",
            coordinateTransformations=(VectorScale(type="scale", scale=(2, 2, 2, 2)),),
        ),
        Dataset(
            path="1",
            coordinateTransformations=(VectorScale(type="scale", scale=(2, 2, 1, 2)),),
        ),
    )
    with pytest.raises(
        ValidationError,
        match=re.escape(
            "Dataset 0 has a lower resolution (scales = [2.0, 2.0, 2.0, 2.0]) "
            "than dataset 1 (scales = [2.0, 2.0, 1.0, 2.0])"
        ),
    ):
        Multiscale(
            axes=axes,
            datasets=datasets,
        )


@pytest.mark.skip
def test_multiscale_group_datasets_exist(
    default_multiscale: Multiscale,
) -> None:
    group_attrs = ImageAttrs(multiscales=(default_multiscale,))
    good_items: dict[str, AnyArraySpec] = {
        d.path: ArraySpec(
            shape=(1, 1, 1, 1),
            dtype="uint8",
            chunks=(1, 1, 1, 1),
        )
        for d in default_multiscale.datasets
    }
    Image(attributes=group_attrs, members=good_items)

    bad_items: dict[str, AnyArraySpec] = {
        d.path + "x": ArraySpec(
            shape=(1, 1, 1, 1),
            dtype="uint8",
            chunks=(1, 1, 1, 1),
        )
        for d in default_multiscale.datasets
    }

    with pytest.raises(
        ValidationError,
        match="array with that name was found in the hierarchy",
    ):
        Image(attributes=group_attrs, members=bad_items)


def test_multiscale_group_datasets_ndim() -> None:
    """
    Test that creating a Image with arrays with mismatched shapes raises
    an exception

    > The length of "axes" ... MUST be equal to the dimensionality of the zarr arrays
    > storing the image data
    """
    true_ndim = 2
    bad_ndim = 3
    match = (
        "Length of arrays (got len(array_specs)=3) must be the same as "
        "length of paths (got len(paths)=2)"
    )
    with pytest.raises(ValueError, match=re.escape(match)):
        Image.new(
            array_specs=[
                ArraySpec(shape=(10,), chunks=(10,), dtype="uint8")
                for _ in range(bad_ndim)
            ],
            paths=[str(i) for i in range(true_ndim)],
            axes=(Axis(name="x", type="space"), Axis(name="y", type="space")),
            scales=((1, 1), (2, 2)),
            translations=((0, 0), (0.5, 0.5)),
        )


def test_multiscale_group_missing_arrays() -> None:
    """
    Test that creating a multiscale group fails when an expected Zarr array is missing
    """
    arrays = np.zeros((10, 10)), np.zeros((5, 5))
    array_names = ("s0", "s1")
    group_model = Image.new(
        array_specs=[ArraySpec.from_array(a) for a in arrays],
        axes=(Axis(name="x", type="space"), Axis(name="y", type="space")),
        paths=array_names,
        scales=((1, 1), (2, 2)),
        translations=((0, 0), (0.5, 0.5)),
    )
    # remove an array, then re-create the model
    assert group_model.members is not None
    group_model_broken = group_model.model_copy(
        update={"members": {array_names[0]: group_model.members[array_names[0]]}}
    )
    with pytest.raises(
        ValidationError,
        match=(
            "The multiscale metadata references an array that does not exist in this "
        ),
    ):
        Image(**group_model_broken.model_dump())


def test_multiscale_group_ectopic_group() -> None:
    """
    Test that creating a multiscale group fails when an expected Zarr array
    is actually a group
    """
    arrays = np.zeros((10, 10)), np.zeros((5, 5))
    array_names = ("s0", "s1")
    group_model = Image.new(
        array_specs=[ArraySpec.from_array(a) for a in arrays],
        axes=(Axis(name="x", type="space"), Axis(name="y", type="space")),
        paths=array_names,
        scales=((1, 1), (2, 2)),
        translations=((0, 0), (0.5, 0.5)),
    )
    # remove an array, then re-create the model
    group_model_broken = group_model.model_copy(
        update={"members": {array_names[0]: GroupSpec()}}
    )
    with pytest.raises(
        ValidationError,
        match=re.escape(f"The node at {array_names[0]} is a group, not an array."),
    ):
        Image(**group_model_broken.model_dump())


def test_from_zarr_missing_metadata(
    store: Store,
    request: pytest.FixtureRequest,
) -> None:
    group_model: AnyGroupSpec = GroupSpec()
    group = group_model.to_zarr(store, path="test")
    # store_path = store.path if hasattr(store, "path") else ""
    match = "multiscales\n  Field required"
    with pytest.raises(ValidationError, match=match):
        Image.from_zarr(group)


def test_from_zarr_missing_array(store: Store) -> None:
    """
    Test that creating a multiscale Group fails when an expected Zarr array is missing
    or is a group instead of an array
    """
    arrays = np.zeros((10, 10)), np.zeros((5, 5))
    group_path = "broken"
    arrays_names = ("s0", "s1")
    group_model = Image.new(
        array_specs=[ArraySpec.from_array(a) for a in arrays],
        axes=(Axis(name="x", type="space"), Axis(name="y", type="space")),
        paths=arrays_names,
        scales=((1, 1), (2, 2)),
        translations=((0, 0), (0.5, 0.5)),
    )

    # make an untyped model, and remove an array before serializing
    removed_array_path = arrays_names[0]
    model_dict = group_model.model_dump(exclude={"members": {removed_array_path: True}})
    broken_group = GroupSpec(**model_dict).to_zarr(store=store, path=group_path)
    match = "Expected to find an array at s0, but no array was found there"
    with pytest.raises(ValueError, match=match):
        Image.from_zarr(broken_group)


def test_from_zarr_ectopic_group(store: Store) -> None:
    """
    Test that creating a multiscale Group fails when an expected Zarr array is missing
    or is a group instead of an array
    """
    arrays = np.zeros((10, 10)), np.zeros((5, 5))
    group_path = "broken"
    arrays_names = ("s0", "s1")
    group_model = Image.new(
        array_specs=[ArraySpec.from_array(a) for a in arrays],
        axes=(Axis(name="x", type="space"), Axis(name="y", type="space")),
        paths=arrays_names,
        scales=((1, 1), (2, 2)),
        translations=((0, 0), (0.5, 0.5)),
    )

    # make an untyped model, and remove an array before serializing
    removed_array_path = arrays_names[0]
    model_dict = group_model.model_dump(exclude={"members": {removed_array_path: True}})
    broken_group = GroupSpec(**model_dict).to_zarr(store=store, path=group_path)

    # put a group where the array should be
    broken_group.create_group(removed_array_path)
    match = "Expected to find an array at s0, but no array was found there"
    with pytest.raises(ValueError, match=match):
        Image.from_zarr(broken_group)


@pytest.mark.skip
def test_hashable(default_multiscale: Multiscale) -> None:
    """
    Test that `Multiscale` can be hashed
    """
    assert set(default_multiscale) == set(default_multiscale)



from typing import Annotated

from ome_zarr_models.base import BaseAttrs
from ome_zarr_models.common.validation import RGBHexConstraint

__all__ = ["Channel", "Omero", "Window"]


class Window(BaseAttrs):
    """
    A single window.
    """

    max: float
    min: float
    start: float
    end: float


class Channel(BaseAttrs):
    """
    A single omero channel.
    """

    color: Annotated[str, RGBHexConstraint]
    window: Window


class Omero(BaseAttrs):
    """
    omero model.
    """

    channels: list[Channel]



import re
from pathlib import Path

import pytest
import zarr
from zarr.abc.store import Store

import ome_zarr_models.v04
import ome_zarr_models.v05
from ome_zarr_models import open_ome_zarr
from tests.conftest import get_examples_path
from tests.v05.test_image import make_valid_image_group


def test_load_ome_zarr_group() -> None:
    hcs_group = zarr.open_group(
        get_examples_path(version="0.4") / "hcs_example.ome.zarr",
        mode="r",
        zarr_format=2,
    )
    ome_zarr_group = open_ome_zarr(hcs_group)

    assert isinstance(ome_zarr_group, ome_zarr_models.v04.HCS)
    assert ome_zarr_group.ome_zarr_version == "0.4"


def test_load_ome_zarr_group_v05_image_label(store: Store) -> None:
    # Check that images and image-labels are distinguished correctly
    image_group = make_valid_image_group(store)
    ome_zarr_group = open_ome_zarr(image_group)

    assert isinstance(ome_zarr_group, ome_zarr_models.v05.Image)

    # Turn into an image labels by adding the image-label metadata
    attrs = image_group.attrs.asdict()
    attrs["ome"]["image-label"] = {}  # type: ignore[index]
    image_group.update_attributes(attrs)
    ome_zarr_group = open_ome_zarr(image_group)
    assert isinstance(ome_zarr_group, ome_zarr_models.v05.ImageLabel)


def test_load_ome_zarr_group_bad(tmp_path: Path) -> None:
    hcs_group = zarr.create_group(tmp_path / "test")
    with pytest.raises(
        RuntimeError,
        match=re.escape(
            f"Could not successfully validate <Group file://{tmp_path / 'test'}> "
        ),
    ):
        open_ome_zarr(hcs_group)


@pytest.mark.vcr
def test_load_remote_data() -> None:
    grp = open_ome_zarr(
        "https://uk1s3.embassy.ebi.ac.uk/idr/zarr/v0.5/idr0066/ExpA_VIP_ASLM_on.zarr",
        version="0.5",
    )
    assert isinstance(grp, ome_zarr_models.v05.Image)



"""
Private utilities.
"""

from __future__ import annotations

from collections import Counter
from dataclasses import MISSING, fields, is_dataclass
from typing import TYPE_CHECKING, Any, TypeVar

import pydantic
import pydantic_zarr.v2
import pydantic_zarr.v3
from pydantic import create_model

from ome_zarr_models.base import BaseAttrsv2, BaseAttrsv3
from ome_zarr_models.common.validation import (
    check_array_path,
    check_group_path,
)

if TYPE_CHECKING:
    from collections.abc import Iterable

    import zarr
    from zarr.abc.store import Store

    from ome_zarr_models._v06.base import BaseGroupv06
    from ome_zarr_models.v04.base import BaseGroupv04
    from ome_zarr_models.v05.base import BaseGroupv05

TBaseGroupv2 = TypeVar("TBaseGroupv2", bound="BaseGroupv04[Any]")
TAttrsv2 = TypeVar("TAttrsv2", bound=BaseAttrsv2)


def _from_zarr_v2(
    group: zarr.Group,
    group_cls: type[TBaseGroupv2],
    attrs_cls: type[TAttrsv2],
) -> TBaseGroupv2:
    """
    Create a GroupSpec from a potentially unlistable Zarr group.

    This uses methods on the attribute class to get required and optional
    paths to ararys and groups, and then manually constructs the GroupSpec
    from those paths.

    Parameters
    ----------
    group :
        Zarr group to create GroupSpec from.
    group_cls :
        Class of the Group to return.
    attrs_cls :
        Attributes class.
    """
    # on unlistable storage backends, the members of this group will be {}
    group_spec_in: pydantic_zarr.v2.AnyGroupSpec
    group_spec_in = pydantic_zarr.v2.GroupSpec.from_zarr(group, depth=0)
    attributes = attrs_cls.model_validate(group_spec_in.attributes)

    members_tree_flat: dict[
        str, pydantic_zarr.v2.AnyGroupSpec | pydantic_zarr.v2.AnyArraySpec
    ] = {}

    # Required array paths
    for array_path in attrs_cls.get_array_paths(attributes):
        array_spec = check_array_path(group, array_path, expected_zarr_version=2)
        members_tree_flat["/" + array_path] = array_spec

    # Optional array paths
    for array_path in attrs_cls.get_optional_array_paths(attributes):
        try:
            array_spec = check_array_path(group, array_path, expected_zarr_version=2)
        except ValueError:
            continue
        members_tree_flat["/" + array_path] = array_spec

    # Required group paths
    required_groups = attrs_cls.get_group_paths(attributes)
    for group_path in required_groups:
        check_group_path(group, group_path, expected_zarr_version=2)
        group_flat = required_groups[group_path].from_zarr(group[group_path]).to_flat()  # type: ignore[arg-type]
        for path in group_flat:
            members_tree_flat["/" + group_path + path] = group_flat[path]

    # Optional group paths
    optional_groups = attrs_cls.get_optional_group_paths(attributes)
    for group_path in optional_groups:
        try:
            check_group_path(group, group_path, expected_zarr_version=2)
        except FileNotFoundError:
            continue
        group_flat = optional_groups[group_path].from_zarr(group[group_path]).to_flat()  # type: ignore[arg-type]
        for path in group_flat:
            members_tree_flat["/" + group_path + path] = group_flat[path]

    members_normalized: pydantic_zarr.v2.AnyGroupSpec = (
        pydantic_zarr.v2.GroupSpec.from_flat(members_tree_flat)
    )
    return group_cls(members=members_normalized.members, attributes=attributes)


TBaseGroupv3 = TypeVar("TBaseGroupv3", bound="BaseGroupv05[Any] | BaseGroupv06[Any]")
TAttrsv3 = TypeVar("TAttrsv3", bound=BaseAttrsv3)


def _from_zarr_v3(
    group: zarr.Group,
    group_cls: type[TBaseGroupv3],
    attrs_cls: type[TAttrsv3],
) -> TBaseGroupv3:
    """
    Create a GroupSpec from a potentially unlistable Zarr group.

    This uses methods on the attribute class to get required and optional
    paths to ararys and groups, and then manually constructs the GroupSpec
    from those paths.

    Parameters
    ----------
    group :
        Zarr group to create GroupSpec from.
    group_cls :
        Class of the Group to return.
    attrs_cls :
        Attributes class.
    """
    # on unlistable storage backends, the members of this group will be {}
    group_spec_in: pydantic_zarr.v3.AnyGroupSpec
    group_spec_in = pydantic_zarr.v3.GroupSpec.from_zarr(group, depth=0)
    ome_attributes = attrs_cls.model_validate(group.attrs.asdict()["ome"])

    members_tree_flat: dict[
        str, pydantic_zarr.v3.AnyGroupSpec | pydantic_zarr.v3.AnyArraySpec
    ] = {}

    # Required array paths
    for array_path in attrs_cls.get_array_paths(ome_attributes):
        array_spec = check_array_path(group, array_path, expected_zarr_version=3)
        members_tree_flat["/" + array_path] = array_spec

    # Optional array paths
    for array_path in attrs_cls.get_optional_array_paths(ome_attributes):
        try:
            array_spec = check_array_path(group, array_path, expected_zarr_version=3)
        except ValueError:
            continue
        members_tree_flat["/" + array_path] = array_spec

    # Required group paths
    required_groups = attrs_cls.get_group_paths(ome_attributes)
    for group_path in required_groups:
        check_group_path(group, group_path, expected_zarr_version=3)
        group_flat = required_groups[group_path].from_zarr(group[group_path]).to_flat()  # type: ignore[arg-type]
        for path in group_flat:
            members_tree_flat["/" + group_path + path] = group_flat[path]

    # Optional group paths
    optional_groups = attrs_cls.get_optional_group_paths(ome_attributes)
    for group_path in optional_groups:
        try:
            check_group_path(group, group_path, expected_zarr_version=3)
        except FileNotFoundError:
            continue
        group_flat = optional_groups[group_path].from_zarr(group[group_path]).to_flat()  # type: ignore[arg-type]
        for path in group_flat:
            members_tree_flat["/" + group_path + path] = group_flat[path]

    members_normalized: pydantic_zarr.v3.AnyGroupSpec
    members_normalized = pydantic_zarr.v3.GroupSpec.from_flat(members_tree_flat)
    return group_cls(  # type: ignore[return-value]
        members=members_normalized.members, attributes=group_spec_in.attributes
    )


def get_store_path(store: Store) -> str:
    """
    Get a path from a zarr store
    """
    if hasattr(store, "path"):
        return store.path  # type: ignore[no-any-return]

    return ""


T = TypeVar("T")


def duplicates(values: Iterable[T]) -> dict[T, int]:
    """
    Takes a sequence of hashable elements and returns a dict where the keys are the
    elements of the input that occurred at least once, and the values are the
    frequencies of those elements.
    """
    counts = Counter(values)
    return {k: v for k, v in counts.items() if v > 1}


def dataclass_to_pydantic(dataclass_type: type) -> type[pydantic.BaseModel]:
    """Convert a dataclass to a Pydantic model.

    Parameters
    ----------
    dataclass_type : type
        The dataclass to convert to a Pydantic model.

    Returns
    -------
    type[pydantic.BaseModel] a Pydantic model class.
    """
    if not is_dataclass(dataclass_type):
        raise TypeError(f"{dataclass_type} is not a dataclass")

    field_definitions = {}
    for _field in fields(dataclass_type):
        if _field.default is not MISSING:
            # Default value is provided
            field_definitions[_field.name] = (_field.type, _field.default)
        elif _field.default_factory is not MISSING:
            # Default factory is provided
            field_definitions[_field.name] = (_field.type, _field.default_factory())
        else:
            # No default value
            field_definitions[_field.name] = (_field.type, Ellipsis)

    return create_model(dataclass_type.__name__, **field_definitions)  # type: ignore[no-any-return, call-overload]



from ome_zarr_models.v04.base import BaseGroupv04
from ome_zarr_models.v04.hcs import HCS
from ome_zarr_models.v04.image import Image
from ome_zarr_models.v04.image_label import ImageLabel
from ome_zarr_models.v04.labels import Labels
from ome_zarr_models.v04.well import Well

__all__ = ["HCS", "BaseGroupv04", "Image", "ImageLabel", "Labels", "Well"]



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Literal

from pydantic import BaseModel, ConfigDict

if TYPE_CHECKING:
    import pydantic_zarr.v2
    import pydantic_zarr.v3


class BaseAttrs(BaseModel):
    """
    The base pydantic model for all metadata classes
    """

    model_config = ConfigDict(
        validate_assignment=True,
        extra="allow",
        # This allows fields with aliases to be populated by either
        # their alias or class attribute name
        #
        # We use this so we can handle e.g., the "bioformats2raw.version"
        # key - names in Python can't contain a "."
        populate_by_name=True,
        frozen=True,
    )


class BaseAttrsv2(BaseAttrs):
    """
    Base attribute model for Zarr v2 groups (ie OME-Zarr 0.4).
    """

    def get_array_paths(self) -> list[str]:
        """
        Get a list of all array paths expected and required to live in the Group
        with these attributes.
        """
        return []

    def get_optional_array_paths(self) -> list[str]:
        """
        Get a list of all array paths expected but not required to live in the Group
        with these attributes.
        """
        return []

    def get_group_paths(self) -> dict[str, type[pydantic_zarr.v2.AnyGroupSpec]]:
        """
        Get a list of all group paths expected and required to live in the Group
        with these attributes.

        Must return a dictionary mapping paths to their GroupSpec class.
        """
        return {}

    def get_optional_group_paths(
        self,
    ) -> dict[str, type[pydantic_zarr.v2.AnyGroupSpec]]:
        """
        Get a list of all group paths expected but not required to live in the Group
        with these attributes.

        Must return a dictionary mapping paths to their GroupSpec class.
        """
        return {}


class BaseAttrsv3(BaseAttrs):
    """
    Base attribute model for Zarr v3 groups (ie OME-Zarr 0.5+).
    """

    def get_array_paths(self) -> list[str]:
        """
        Get a list of all array paths expected and required to live in the Group
        with these attributes.
        """
        return []

    def get_optional_array_paths(self) -> list[str]:
        """
        Get a list of all array paths expected but not required to live in the Group
        with these attributes.
        """
        return []

    def get_group_paths(self) -> dict[str, type[pydantic_zarr.v3.AnyGroupSpec]]:
        """
        Get a list of all group paths expected and required to live in the Group
        with these attributes.

        Must return a dictionary mapping paths to their GroupSpec class.
        """
        return {}

    def get_optional_group_paths(
        self,
    ) -> dict[str, type[pydantic_zarr.v3.AnyGroupSpec]]:
        """
        Get a list of all group paths expected but not required to live in the Group
        with these attributes.

        Must return a dictionary mapping paths to their GroupSpec class.
        """
        return {}


class BaseGroup(ABC):
    """
    Base class for all OME-Zarr groups.
    """

    @property
    @abstractmethod
    def ome_zarr_version(self) -> Literal["0.4", "0.5", "0.6"]:
        """
        Version of the OME-Zarr specification that this group corresponds to.
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def ome_attributes(self) -> BaseAttrs:
        """
        OME attributes.
        """
        raise NotImplementedError



"""
General tests.
"""

import re

import pytest
from pydantic import ValidationError
from zarr.abc.store import Store

from ome_zarr_models.v05.image import Image
from tests.v05.conftest import json_to_zarr_group


def test_no_ome_version_fails(store: Store) -> None:
    zarr_group = json_to_zarr_group(
        json_fname="image_no_version_example.json", store=store
    )
    zarr_group.create_array(
        "0",
        shape=(1, 1, 1, 1, 1),
        dtype="uint8",
    )
    zarr_group.create_array("1", shape=(1, 1, 1, 1, 1), dtype="uint8")
    zarr_group.create_array("2", shape=(1, 1, 1, 1, 1), dtype="uint8")
    with pytest.raises(ValidationError, match=re.escape("version")):
        Image.from_zarr(zarr_group)



from typing import Literal

from pydantic import Field, JsonValue

from ome_zarr_models.base import BaseAttrs
from ome_zarr_models.v04.plate import Plate


class BioFormats2RawAttrs(BaseAttrs):
    """
    A model of the attributes contained in a bioformats2raw zarr group.
    """

    bioformats2raw_layout: Literal[3] = Field(..., alias="bioformats2raw.layout")
    plate: Plate | None = None
    series: JsonValue | None = None



from ome_zarr_models.v05.hcs import HCS
from ome_zarr_models.v05.image import Image
from ome_zarr_models.v05.image_label import ImageLabel
from ome_zarr_models.v05.labels import Labels
from ome_zarr_models.v05.well import Well

__all__ = ["HCS", "Image", "ImageLabel", "Labels", "Well"]



from typing import Self

# Import needed for pydantic type resolution
import pydantic_zarr  # noqa: F401
import zarr
from pydantic import Field

from ome_zarr_models.v05.base import BaseGroupv05, BaseOMEAttrs
from ome_zarr_models.v05.image import Image
from ome_zarr_models.v05.image_label_types import Label
from ome_zarr_models.v05.multiscales import Multiscale

__all__ = ["ImageLabel", "ImageLabelAttrs"]


class ImageLabelAttrs(BaseOMEAttrs):
    """
    Attributes for an image label object.
    """

    image_label: Label | None = Field(alias="image-label", default=None)
    multiscales: list[Multiscale]


class ImageLabel(
    BaseGroupv05[ImageLabelAttrs],
):
    """
    An OME-Zarr image label dataset.
    """

    @classmethod
    def from_zarr(cls, group: zarr.Group) -> Self:  # type: ignore[override]
        """
        Create an instance of an OME-Zarr image from a `zarr.Group`.

        Parameters
        ----------
        group : zarr.Group
            A Zarr group that has valid OME-Zarr image label metadata.
        """
        # Use Image.from_zarr() to validate multiscale metadata
        image = Image.from_zarr(group)
        return cls(attributes=image.attributes.model_dump(), members=image.members)



"""
For reference, see the [image label section of the OME-Zarr specification](https://ngff.openmicroscopy.org/0.4/index.html#label-md).
"""

from __future__ import annotations

from typing import Annotated, Self

from pydantic import Field, field_validator, model_validator

from ome_zarr_models._utils import duplicates
from ome_zarr_models.base import BaseAttrs

__all__ = ["RGBA", "Color", "LabelBase", "Property", "Source", "Uint8"]

Uint8 = Annotated[int, Field(strict=True, ge=0, le=255)]
RGBA = tuple[Uint8, Uint8, Uint8, Uint8]


class Color(BaseAttrs):
    """
    A label value and RGBA.
    """

    label_value: int = Field(..., alias="label-value")
    rgba: RGBA | None


class Source(BaseAttrs):
    """
    Source data for the labels.
    """

    # TODO: add validation that this path resolves to a zarr image group
    image: str | None = Field(
        default="../../", description="Relative path to a Zarr group of a key image."
    )


class Property(BaseAttrs):
    """
    A single property.
    """

    label_value: int = Field(..., alias="label-value")


class LabelBase(BaseAttrs):
    """
    Base class for image-label metadata.
    """

    # TODO: validate
    # "All the values under the label-value (of colors) key MUST be unique."
    colors: tuple[Color, ...] | None = None
    properties: tuple[Property, ...] | None = None
    source: Source | None = None
    version: str | None = None

    @model_validator(mode="after")
    def _check_label_values(self) -> Self:
        """
        Check that label_values are consistent across properties and colors
        """
        if self.colors is not None and self.properties is not None:
            prop_label_value = [prop.label_value for prop in self.properties]
            color_label_value = [color.label_value for color in self.colors]

            prop_label_value_set = set(prop_label_value)
            color_label_value_set = set(color_label_value)
            if color_label_value_set != prop_label_value_set:
                msg = (
                    "Inconsistent `label_value` attributes in "
                    "`colors` and `properties`."
                    "The `properties` attributes have "
                    f"`label_values` {prop_label_value}, "
                    "The `colors` attributes have "
                    f"`label_values` {color_label_value}, "
                )
                raise ValueError(msg)
        return self

    @field_validator("colors", mode="after")
    def _parse_colors(cls, colors: tuple[Color, ...]) -> tuple[Color, ...]:
        """
        Check that color label values are unique.
        """
        # if colors is None:
        #    msg = (
        #        "The field `colors` is `None`. `colors` should be a list of "
        #        "label descriptors."
        #    )
        #    warnings.warn(msg, stacklevel=1)
        dupes = duplicates(x.label_value for x in colors)
        if len(dupes) > 0:
            msg = (
                f"Duplicated label-value: {tuple(dupes.keys())}."
                "label-values must be unique across elements of `colors`."
            )
            raise ValueError(msg)

        return colors



from typing import Generic, Literal, Self, TypeVar

import zarr
from pydantic_zarr.v2 import GroupSpec, TBaseItem

from ome_zarr_models.base import BaseAttrsv2, BaseGroup

T = TypeVar("T", bound=BaseAttrsv2)


class BaseGroupv04(BaseGroup, GroupSpec[T, TBaseItem], Generic[T]):
    """
    Base class for all v0.4 OME-Zarr groups.
    """

    @classmethod
    def from_zarr(cls, group: zarr.Group) -> Self:  # type: ignore[override]
        """
        Create an OME-Zarr model from a `zarr.Group`.

        Parameters
        ----------
        group : zarr.Group
            A Zarr group that has valid OME-Zarr image metadata.
        """
        return super().from_zarr(group)

    @property
    def ome_zarr_version(self) -> Literal["0.4"]:
        """
        OME-Zarr version.
        """
        return "0.4"

    @property
    def ome_attributes(self) -> T:
        """
        OME attributes.
        """
        return self.attributes



from zarr.abc.store import Store

from ome_zarr_models.v05.axes import Axis
from ome_zarr_models.v05.coordinate_transformations import VectorScale
from ome_zarr_models.v05.image_label import ImageLabel, ImageLabelAttrs
from ome_zarr_models.v05.image_label_types import Color, Label, Source
from ome_zarr_models.v05.multiscales import Dataset, Multiscale
from tests.v05.conftest import json_to_zarr_group


def test_image_label(store: Store) -> None:
    zarr_group = json_to_zarr_group(json_fname="image_label_example.json", store=store)
    zarr_group.create_array(
        "0",
        shape=(1, 1, 1, 1, 1),
        dtype="uint8",
        dimension_names=["t", "c", "z", "y", "x"],
    )
    zarr_group.create_array(
        "1",
        shape=(1, 1, 1, 1, 1),
        dtype="uint8",
        dimension_names=["t", "c", "z", "y", "x"],
    )
    zarr_group.create_array(
        "2",
        shape=(1, 1, 1, 1, 1),
        dtype="uint8",
        dimension_names=["t", "c", "z", "y", "x"],
    )
    ome_group = ImageLabel.from_zarr(zarr_group)
    assert ome_group.attributes.ome == ImageLabelAttrs(
        image_label=Label(
            colors=(
                Color(label_value=0, rgba=(0, 0, 128, 128)),
                Color(label_value=1, rgba=(0, 128, 0, 128)),
            ),
            properties=[
                {
                    "label_value": 0,
                    "area (pixels)": 1200,
                    "class": "intercellular space",
                },
                {
                    "label_value": 1,
                    "area (pixels)": 1650,
                    "class": "cell",
                    "cell type": "neuron",
                },
            ],
            source=Source(image="../../"),
            version=None,
        ),
        multiscales=[
            Multiscale(
                axes=[
                    Axis(name="t", type="time", unit="millisecond"),
                    Axis(name="c", type="channel", unit=None),
                    Axis(name="z", type="space", unit="micrometer"),
                    Axis(name="y", type="space", unit="micrometer"),
                    Axis(name="x", type="space", unit="micrometer"),
                ],
                datasets=(
                    Dataset(
                        path="0",
                        coordinateTransformations=(
                            VectorScale(type="scale", scale=[1.0, 1.0, 0.5, 0.5, 0.5]),
                        ),
                    ),
                    Dataset(
                        path="1",
                        coordinateTransformations=(
                            VectorScale(type="scale", scale=[1.0, 1.0, 1.0, 1.0, 1.0]),
                        ),
                    ),
                    Dataset(
                        path="2",
                        coordinateTransformations=(
                            VectorScale(type="scale", scale=[1.0, 1.0, 2.0, 2.0, 2.0]),
                        ),
                    ),
                ),
                coordinateTransformations=(
                    VectorScale(type="scale", scale=[0.1, 1.0, 1.0, 1.0, 1.0]),
                ),
                metadata={
                    "description": (
                        "the fields in metadata depend on the downscaling "
                        "implementation. Here, the parameters passed to the skimage "
                        "function are given"
                    ),
                    "method": "skimage.transform.pyramid_gaussian",
                    "version": "0.16.1",
                    "args": "[true]",
                    "kwargs": {"multichannel": True},
                },
                name="example",
                type="gaussian",
            )
        ],
        version="0.5",
    )



import re

import pytest
import zarr
from pydantic import ValidationError
from zarr.abc.store import Store

from ome_zarr_models.v05.axes import Axis
from ome_zarr_models.v05.coordinate_transformations import VectorScale
from ome_zarr_models.v05.image import Image, ImageAttrs
from ome_zarr_models.v05.labels import LabelsAttrs
from ome_zarr_models.v05.multiscales import Dataset, Multiscale
from tests.v05.conftest import json_to_dict, json_to_zarr_group


def make_valid_image_group(store: Store) -> zarr.Group:
    zarr_group = json_to_zarr_group(json_fname="image_example.json", store=store)
    zarr_group.create_array(
        "0",
        shape=(1, 1, 1, 1, 1),
        dtype="uint8",
        dimension_names=["t", "c", "z", "y", "x"],
    )
    zarr_group.create_array(
        "1",
        shape=(1, 1, 1, 1, 1),
        dtype="uint8",
        dimension_names=["t", "c", "z", "y", "x"],
    )
    zarr_group.create_array(
        "2",
        shape=(1, 1, 1, 1, 1),
        dtype="uint8",
        dimension_names=["t", "c", "z", "y", "x"],
    )
    return zarr_group


def test_image(store: Store) -> None:
    zarr_group = make_valid_image_group(store)
    ome_group = Image.from_zarr(zarr_group)

    assert ome_group.attributes.ome == ImageAttrs(
        multiscales=[
            Multiscale(
                axes=[
                    Axis(name="t", type="time", unit="millisecond"),
                    Axis(name="c", type="channel", unit=None),
                    Axis(name="z", type="space", unit="micrometer"),
                    Axis(name="y", type="space", unit="micrometer"),
                    Axis(name="x", type="space", unit="micrometer"),
                ],
                datasets=(
                    Dataset(
                        path="0",
                        coordinateTransformations=(
                            VectorScale(type="scale", scale=[1.0, 1.0, 0.5, 0.5, 0.5]),
                        ),
                    ),
                    Dataset(
                        path="1",
                        coordinateTransformations=(
                            VectorScale(type="scale", scale=[1.0, 1.0, 1.0, 1.0, 1.0]),
                        ),
                    ),
                    Dataset(
                        path="2",
                        coordinateTransformations=(
                            VectorScale(type="scale", scale=[1.0, 1.0, 2.0, 2.0, 2.0]),
                        ),
                    ),
                ),
                coordinateTransformations=(
                    VectorScale(type="scale", scale=[0.1, 1.0, 1.0, 1.0, 1.0]),
                ),
                metadata={
                    "description": "the fields in metadata depend on the downscaling "
                    "implementation. Here, the parameters passed to the "
                    "skimage function are given",
                    "method": "skimage.transform.pyramid_gaussian",
                    "version": "0.16.1",
                    "args": "[true]",
                    "kwargs": {"multichannel": True},
                },
                name="example",
                type="gaussian",
            )
        ],
        version="0.5",
    )


def test_image_no_dim_names(store: Store) -> None:
    zarr_group = json_to_zarr_group(json_fname="image_example.json", store=store)
    zarr_group.create_array(
        "0",
        shape=(1, 1, 1, 1, 1),
        dtype="uint8",
        dimension_names=["t", "c", "z", "y", "x"],
    )
    arr1 = zarr_group.create_array("1", shape=(1, 1, 1, 1, 1), dtype="uint8")
    assert arr1.metadata.dimension_names is None  # type: ignore[union-attr]
    zarr_group.create_array(
        "2",
        shape=(1, 1, 1, 1, 1),
        dtype="uint8",
        dimension_names=["t", "c", "z", "y", "x"],
    )
    with pytest.raises(
        ValidationError,
        match="The array in this group at  '1' has no dimension_names metadata",
    ):
        Image.from_zarr(zarr_group)


def test_image_wrong_dim_names(store: Store) -> None:
    zarr_group = json_to_zarr_group(json_fname="image_example.json", store=store)
    zarr_group.create_array(
        "0",
        shape=(1, 1, 1, 1, 1),
        dtype="uint8",
        dimension_names=["t", "c", "z", "y", "x"],
    )
    zarr_group.create_array(
        "1",
        shape=(1, 1, 1, 1, 1),
        dtype="uint8",
        dimension_names=["t", "c", "z", "x", "y"],
    )
    zarr_group.create_array(
        "2",
        shape=(1, 1, 1, 1, 1),
        dtype="uint8",
        dimension_names=["t", "c", "z", "y", "x"],
    )
    with pytest.raises(
        ValidationError,
        match=re.escape(
            "The multiscale metadata has ('t', 'c', 'z', 'y', 'x') axes names "
            "which does not match the dimension names of the array "
            "found in this group at path '1' (('t', 'c', 'z', 'x', 'y'))"
        ),
    ):
        Image.from_zarr(zarr_group)


def test_image_with_labels(store: Store) -> None:
    zarr_group = json_to_zarr_group(json_fname="image_example.json", store=store)
    zarr_group.create_array(
        "0",
        shape=(1, 1, 1, 1, 1),
        dtype="uint8",
        dimension_names=["t", "c", "z", "y", "x"],
    )
    zarr_group.create_array(
        "1",
        shape=(1, 1, 1, 1, 1),
        dtype="uint8",
        dimension_names=["t", "c", "z", "y", "x"],
    )
    zarr_group.create_array(
        "2",
        shape=(1, 1, 1, 1, 1),
        dtype="uint8",
        dimension_names=["t", "c", "z", "y", "x"],
    )
    # Add labels group
    labels_group = zarr_group.create_group(
        "labels",
        attributes=json_to_dict(json_fname="labels_example.json"),
    )

    with pytest.raises(
        ValueError, match="Label path 'cell_space_segmentation' not found in zarr group"
    ):
        Image.from_zarr(zarr_group)

    # Add image labels group
    image_label_group = labels_group.create_group(
        "cell_space_segmentation",
        attributes=json_to_dict(json_fname="image_label_example.json"),
    )
    image_label_group.create_array(
        "0",
        shape=(1, 1, 1, 1, 1),
        dtype="uint8",
        dimension_names=["t", "c", "z", "y", "x"],
    )
    image_label_group.create_array(
        "1",
        shape=(1, 1, 1, 1, 1),
        dtype="uint8",
        dimension_names=["t", "c", "z", "y", "x"],
    )
    image_label_group.create_array(
        "2",
        shape=(1, 1, 1, 1, 1),
        dtype="uint8",
        dimension_names=["t", "c", "z", "y", "x"],
    )
    image = Image.from_zarr(zarr_group)
    assert image.labels is not None
    assert image.labels.attributes.ome == LabelsAttrs(
        version="0.5", labels=["cell_space_segmentation"]
    )


def test_image_with_labels_mismatch_multiscales(store: Store) -> None:
    zarr_group = json_to_zarr_group(json_fname="image_example.json", store=store)
    zarr_group.create_array(
        "0",
        shape=(1, 1, 1, 1, 1),
        dtype="uint8",
        dimension_names=["t", "c", "z", "y", "x"],
    )
    zarr_group.create_array(
        "1",
        shape=(1, 1, 1, 1, 1),
        dtype="uint8",
        dimension_names=["t", "c", "z", "y", "x"],
    )
    zarr_group.create_array(
        "2",
        shape=(1, 1, 1, 1, 1),
        dtype="uint8",
        dimension_names=["t", "c", "z", "y", "x"],
    )
    # Add labels group
    labels_group = zarr_group.create_group(
        "labels",
        attributes=json_to_dict(json_fname="labels_example.json"),
    )
    # Add image labels group
    image_label_group = labels_group.create_group(
        "cell_space_segmentation",
        attributes=json_to_dict(json_fname="labels_image_example.json"),
    )
    image_label_group.create_array(
        "0",
        shape=(1, 1, 1, 1, 1),
        dtype="uint8",
        dimension_names=["t", "c", "z", "y", "x"],
    )
    with pytest.raises(
        RuntimeError,
        match=re.escape(
            "Number of image label multiscale levels (1) doesn't match "
            "number of image multiscale levels (3)."
        ),
    ):
        Image.from_zarr(zarr_group)



from zarr.abc.store import Store

from ome_zarr_models._v06.hcs import HCS, HCSAttrs
from ome_zarr_models._v06.plate import Acquisition, Column, Plate, Row, WellInPlate
from tests._v06.conftest import json_to_zarr_group


def test_hcs(store: Store) -> None:
    zarr_group = json_to_zarr_group(json_fname="hcs_example.json", store=store)
    ome_group = HCS.from_zarr(zarr_group)
    assert ome_group.attributes.ome == HCSAttrs(
        plate=Plate(
            acquisitions=[
                Acquisition(
                    id=1,
                    name="Meas_01(2012-07-31_10-41-12)",
                    maximumfieldcount=2,
                    description=None,
                    starttime=1343731272000,
                    endtime=None,
                ),
                Acquisition(
                    id=2,
                    name="Meas_02(201207-31_11-56-41)",
                    maximumfieldcount=2,
                    description=None,
                    starttime=1343735801000,
                    endtime=None,
                ),
            ],
            columns=[Column(name="1"), Column(name="2"), Column(name="3")],
            field_count=4,
            name="test",
            rows=[Row(name="A"), Row(name="B")],
            wells=[
                WellInPlate(path="A/1", rowIndex=0, columnIndex=0),
                WellInPlate(path="A/2", rowIndex=0, columnIndex=1),
                WellInPlate(path="A/3", rowIndex=0, columnIndex=2),
                WellInPlate(path="B/1", rowIndex=1, columnIndex=0),
                WellInPlate(path="B/2", rowIndex=1, columnIndex=1),
                WellInPlate(path="B/3", rowIndex=1, columnIndex=2),
            ],
            version="0.6",
        ),
        version="0.6",
    )
    well_groups = list(ome_group.well_groups)
    assert len(well_groups) == 0



from __future__ import annotations

from typing import TYPE_CHECKING, cast

import pytest
import zarr
from zarr.storage import LocalStore

from ome_zarr_models._cli import main

from .conftest import Version, json_to_zarr_group

if TYPE_CHECKING:
    from collections.abc import Mapping, Sequence
    from pathlib import Path
    from typing import Any


def populate_fake_data(
    zarr_group: zarr.Group,
    default_dtype: str = "uint8",
) -> None:
    # Get the ome metadata from the group attributes
    # version 0.4 uses the root attributes, version 0.5 uses the "ome" attribute
    ome_attrs = cast("Mapping[str, Any]", zarr_group.attrs.get("ome", zarr_group.attrs))
    multiscales = ome_attrs.get("multiscales")
    if isinstance(multiscales, list):
        create_multiscales_data(zarr_group, multiscales, default_dtype)
        return

    # TODO?  could support fake data for other node types


def create_multiscales_data(
    zarr_group: zarr.Group,
    multiscales: Sequence[Mapping[str, Any]],
    default_dtype: str = "uint8",
) -> None:
    # Use the first multiscale (most common case)
    for multiscale in multiscales:
        if not (axes := multiscale.get("axes")):
            raise ValueError(
                f"No axes found in multiscale metadata from {zarr_group.store_path}"
            )
        if not (datasets := multiscale.get("datasets")):
            raise ValueError(
                f"No datasets found in multiscale metadata from {zarr_group.store_path}"
            )

        dimension_names = [axis["name"] for axis in axes]
        shape = (1,) * len(dimension_names)
        kwargs = {}
        if zarr_group.metadata.zarr_format >= 3:
            kwargs.update({"dimension_names": dimension_names})

        # Create arrays for each dataset path
        for dataset in datasets:
            if path := dataset.get("path"):
                zarr_group.create_array(
                    path,
                    shape=shape,
                    dtype=default_dtype,
                    **kwargs,  # type: ignore[arg-type]
                )


@pytest.mark.parametrize(
    "version,json_fname",
    [
        ("0.4", "multiscales_example.json"),
        ("0.5", "image_example.json"),
        ("0.5", "image_label_example.json"),
        ("0.5", "plate_example_1.json"),
    ],
)
@pytest.mark.parametrize("cmd", ["validate", "info"])
def test_cli_validate(
    version: Version,
    json_fname: str,
    cmd: str,
    monkeypatch: pytest.MonkeyPatch,
    tmp_path: Path,
    capsys: pytest.CaptureFixture[str],
) -> None:
    """Test the CLI commands."""

    zarr_group = json_to_zarr_group(
        version=version, json_fname=json_fname, store=LocalStore(root=tmp_path)
    )
    populate_fake_data(zarr_group)
    monkeypatch.setattr("sys.argv", ["ome-zarr-models", cmd, str(tmp_path)])
    main()
    if cmd == "validate":
        assert "Valid OME-Zarr" in capsys.readouterr().out


@pytest.mark.parametrize("cmd", ["validate", "info"])
def test_cli_invalid(
    tmp_path: Path,
    cmd: str,
    monkeypatch: pytest.MonkeyPatch,
    capsys: pytest.CaptureFixture[str],
) -> None:
    """Test the CLI with no command."""
    zarr.create_group(tmp_path)
    monkeypatch.setattr("sys.argv", ["ome-zarr-models", cmd, str(tmp_path)])
    with pytest.raises(SystemExit) as excinfo:
        main()
    assert excinfo.value.code == 1
    assert "Invalid OME-Zarr" in capsys.readouterr().out



"""
For reference, see the [coordinate transformations section of the OME-Zarr specification](https://ngff.openmicroscopy.org/0.4/#trafo-md).
"""

from ome_zarr_models.common.coordinate_transformations import (
    Identity,
    PathScale,
    PathTranslation,
    ScaleTransform,
    Transform,
    TranslationTransform,
    VectorScale,
    VectorTransform,
    VectorTranslation,
)

__all__ = [
    "Identity",
    "PathScale",
    "PathTranslation",
    "ScaleTransform",
    "Transform",
    "TranslationTransform",
    "VectorScale",
    "VectorTransform",
    "VectorTranslation",
]



# Import needed for pydantic type resolution
from typing import Self

import pydantic_zarr  # noqa: F401
import zarr

from ome_zarr_models._utils import _from_zarr_v3
from ome_zarr_models.v05.base import BaseGroupv05, BaseOMEAttrs
from ome_zarr_models.v05.image import Image
from ome_zarr_models.v05.well_types import WellMeta

__all__ = ["Well", "WellAttrs"]


class WellAttrs(BaseOMEAttrs):
    """
    Attributes for a well.
    """

    well: WellMeta

    def get_optional_group_paths(self) -> dict[str, type[Image]]:  # type: ignore[override]
        return {im.path: Image for im in self.well.images}


class Well(BaseGroupv05[WellAttrs]):
    """
    An OME-Zarr well dataset.
    """

    @classmethod
    def from_zarr(cls, group: zarr.Group) -> Self:  # type: ignore[override]
        """
        Create an OME-Zarr well model from a `zarr.Group`.

        Parameters
        ----------
        group : zarr.Group
            A Zarr group that has valid OME-Zarr well metadata.
        """
        return _from_zarr_v3(group, cls, WellAttrs)
